"""
Reference library for AIRR schema for Ig/TCR rearrangements
"""
from airr.interface import *
from airr.schema import ValidationError

# versioneer-generated
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
